@@ -588,7 +588,8 @@ static bool discover_descs(struct discovery_op *op, bool *discovering)
 	*discovering = false;
 
 	while ((chrc_data = queue_pop_head(op->pending_chrcs))) {
-		attr = gatt_db_service_add_characteristic(op->cur_svc,
+		attr = gatt_db_service_insert_characteristic(op->cur_svc,
+							chrc_data->value_handle,
 							&chrc_data->uuid, 0,
 							chrc_data->properties,
 							NULL, NULL, NULL);
@@ -679,8 +680,9 @@ static void discover_descs_cb(bool success, uint8_t att_ecode,
 						"handle: 0x%04x, uuid: %s",
 						handle, uuid_str);
 
-		attr = gatt_db_service_add_descriptor(op->cur_svc, &uuid, 0,
-							NULL, NULL, NULL);
+		attr = gatt_db_service_insert_descriptor(op->cur_svc, handle,
+							&uuid, 0, NULL, NULL,
+							NULL);
 		if (!attr)
 			goto failed;
 
@@ -688,13 +690,13 @@ static void discover_descs_cb(bool success, uint8_t att_ecode,
 			goto failed;
 	}
 
+next:
 	if (!discover_descs(op, &discovering))
 		goto failed;
 
 	if (discovering)
 		return;
 
-next:
 	/* Done with the current service */
 	gatt_db_service_set_active(op->cur_svc, true);
 
@@ -1815,6 +1817,19 @@ static bool cancel_prep_write_session(struct bt_gatt_client *client,
 							req, request_unref);
 }
 
+static bool cancel_request(struct request *req)
+{
+	req->removed = true;
+
+	if (req->long_write)
+		return cancel_long_write_req(req->client, req);
+
+	if (req->prep_write)
+		return cancel_prep_write_session(req->client, req);
+
+	return bt_att_cancel(req->client->att, req->att_id);
+}
+
 bool bt_gatt_client_cancel(struct bt_gatt_client *client, unsigned int id)
 {
 	struct request *req;
@@ -1827,35 +1842,7 @@ bool bt_gatt_client_cancel(struct bt_gatt_client *client, unsigned int id)
 	if (!req)
 		return false;
 
-	req->removed = true;
-
-	if (!bt_att_cancel(client->att, req->att_id) && !req->long_write &&
-							!req->prep_write)
-		return false;
-
-	/* If this was a long-write, we need to abort all prepared writes */
-	if (req->long_write)
-		return cancel_long_write_req(client, req);
-
-	if (req->prep_write)
-		return cancel_prep_write_session(client, req);
-
-	return true;
-}
-
-static void cancel_request(void *data)
-{
-	struct request *req = data;
-
-	req->removed = true;
-
-	bt_att_cancel(req->client->att, req->att_id);
-
-	if (req->long_write)
-		cancel_long_write_req(req->client, req);
-
-	if (req->prep_write)
-		cancel_prep_write_session(req->client, req);
+	return cancel_request(req);
 }
 
 bool bt_gatt_client_cancel_all(struct bt_gatt_client *client)
@@ -1863,7 +1850,8 @@ bool bt_gatt_client_cancel_all(struct bt_gatt_client *client)
 	if (!client || !client->att)
 		return false;
 
-	queue_remove_all(client->pending_requests, NULL, NULL, cancel_request);
+	queue_remove_all(client->pending_requests, NULL, NULL,
+					(queue_destroy_func_t) cancel_request);
 
 	if (client->discovery_req) {
 		bt_gatt_request_cancel(client->discovery_req);
@@ -2715,7 +2703,7 @@ static struct request *get_reliable_request(struct bt_gatt_client *client,
 
 unsigned int bt_gatt_client_prepare_write(struct bt_gatt_client *client,
 				unsigned int id, uint16_t value_handle,
-				uint16_t offset, uint8_t *value,
+				uint16_t offset, const uint8_t *value,
 				uint16_t length,
 				bt_gatt_client_write_long_callback_t callback,
 				void *user_data,
